unit uArt;

interface

uses Windows, SysUtils, Classes;

type
  PHead = ^THead;
  THead = array[$00..$20] of DWORD;

  PPaletteInfo = ^TPaletteInfo;
  TPaletteInfo = array[$00..$FF] of DWORD; //256 * 4 byte

  PImageInfo = ^TImageInfo;
  TImageInfo = packed record
    //Image size
    Width: DWORD;
    Height: DWORD;
    //Image data size (in art file)
    Size: DWORD;
    //(-Left, -Top)  offset of the upper left corner of image
    Left, Top: Integer;
    //(DeltaX, DeltaY)  offset of image when changing frames
    //(used for critters' movement)
    DeltaX, DeltaY: Integer;
  end;

  PImageInfoArr = ^TImageInfoArr;
  TImageInfoArr = array[0..$FFFF] of TImageInfo;

  TArtInfo = class
  private
    FStream: TMemoryStream;
    FPictureCount: Integer;
    FFrameCount: Integer;
    FPaletteCount: Integer;
  protected
    procedure LoadInfo;  
  public                                 // Art file structure:
    Head: PHead;                         // Header size is 0x84h
    Palette: array[0..3] of PPaletteInfo;// from 1 to 4 palettes, each is 0x400
    ImageInfo: PImageInfoArr;            // Image infos [PC*FC], each is 0x1C
    Images: Pointer;                     // Images [PC*FC]  size is in info
  public
    destructor Destroy; override;
    procedure Clear;
    //Creating DDB image (for the current desktop)
    function CreateBitmap(ImgIndx, PalIndx: Integer; Mirror: boolean;
      var Bitmap: HBITMAP): boolean;
    //Creating DIB stream of image (256 colors)
    function CreateBitmapStream(ImgIndx, PalIndx: Integer; Mirror: boolean;
      Stream: TMemoryStream): boolean;
    //Replacing image
    procedure ReplaceBitmap(ImgIndx, PalIndx: Integer;
      ReplacePalette: boolean; Stream: TMemoryStream);
    //Loading art file
    procedure LoadFile(const FileName: string);
    //Saving art file
    procedure SaveFile(const FileName: string);
    function IsEmpty: boolean;
    function GetImageInfoPtr(Picture, Frame: Integer): PImageInfo; overload;
    function GetImageInfoPtr(Index: Integer): PImageInfo; overload;
    function GetImagePtr(Picture, Frame: Integer): Pointer; overload;
    function GetImagePtr(Index: Integer): Pointer; overload;
    property PictureCount: Integer read FPictureCount;
    property FrameCount: Integer read FFrameCount;
    property PaletteCount: Integer read FPaletteCount;
  end;

implementation

{ TArtInfo }

procedure TArtInfo.Clear;
var
  i: integer;
begin
  FreeAndNil(FStream);
  Head := nil;
  for i := 0 to 3 do
    Palette[i] := nil;
  ImageInfo := nil;
end;

function TArtInfo.CreateBitmap(ImgIndx, PalIndx: Integer; Mirror: boolean;
  var Bitmap: HBITMAP): boolean;
var
  Stream: TMemoryStream;
  BitmapInfo: PBitmapInfo;
  BitsMem: Pointer;
  DC: HDC;
begin
  Stream := TMemoryStream.Create;
  try
    Result := CreateBitmapStream(ImgIndx, PalIndx, Mirror, Stream);
    if not Result then Exit;
    BitsMem := Stream.Memory;
    Inc(PChar(BitsMem), SizeOf(TBitmapFileHeader));
    BitmapInfo := PBitmapInfo(BitsMem);
    Inc(PChar(BitsMem), SizeOf(TBitmapInfoHeader) + SizeOf(TPaletteInfo));
    DC := GetDC(0);
    if DC = 0 then RaiseLastWin32Error;
    try
      Bitmap := CreateDIBitmap(DC, BitmapInfo^.bmiHeader, CBM_INIT,
        BitsMem, BitmapInfo^, DIB_RGB_COLORS);
      if Bitmap = 0 then RaiseLastWin32Error;
    finally
      ReleaseDC(0, DC);
    end;
  finally
    Stream.Free;
  end;
end;

function TArtInfo.CreateBitmapStream(ImgIndx, PalIndx: Integer; Mirror: boolean;
  Stream: TMemoryStream): boolean;
var
  Flag: boolean;
  i: integer;
  Tmp: Char;
  Cnt, Pt, HSz, NW, EID, W: DWORD;
  ImgInfo: PImageInfo;
  ImgData: Pointer;
  P: Pointer;
begin
  Result := not IsEmpty;
  if IsEmpty then
  begin
    Stream.Size := 0;
    Exit;
  end;
  if (PalIndx < 0) or (PalIndx >= PaletteCount) then PalIndx := 0;
  ImgInfo := GetImageInfoPtr(ImgIndx);
  NW := (ImgInfo.Width + 3) and $FFFFFFFC; //Making width multiple of 4
  HSz := SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader)
    + SizeOf(TPaletteInfo);
  Stream.Size := HSz + ImgInfo.Height * NW;
  //TBitmapFileHeader
  P := Stream.Memory;
  PBitmapFileHeader(P).bfType := $4D42;
  PBitmapFileHeader(P).bfSize := Stream.Size;
  PBitmapFileHeader(P).bfReserved1 := 0;
  PBitmapFileHeader(P).bfReserved2 := 0;
  PBitmapFileHeader(P).bfOffBits := HSz;
  //TBitmapInfoHeader
  Inc(PChar(P), SizeOf(TBitmapFileHeader));
  FillChar(P^, SizeOf(TBitmapInfoHeader), 0);
  PBitmapInfoHeader(P).biSize := SizeOf(TBitmapInfoHeader);
  PBitmapInfoHeader(P).biWidth := ImgInfo.Width;
  PBitmapInfoHeader(P).biHeight := ImgInfo.Height;
  PBitmapInfoHeader(P).biPlanes := 1;
  PBitmapInfoHeader(P).biBitCount := 8;
  //Palette
  Inc(PChar(P), SizeOf(TBitmapInfoHeader));
  Move(Palette[PalIndx]^, P^, SizeOf(TPaletteInfo));
  //Image
  // We should remember that BMP files have the reverse line order
  // and the length of each line is multiple of 4 (the rest is ignored)
  Inc(PChar(P), SizeOf(TPaletteInfo) + (ImgInfo.Height - 1) * NW);
  ImgData := GetImagePtr(ImgIndx);
  //!!! There's a suspicion that it's incorrect to identify a data type this way
  // (Need to look for a flag)
  if ImgInfo.Width * ImgInfo.Height = ImgInfo.Size then //RAW ?
    begin
      for i := 0 to ImgInfo.Height - 1 do
      begin
        Move(ImgData^, P^, ImgInfo.Width);
        Dec(PChar(P), NW);
        Inc(PChar(ImgData), ImgInfo.Width);
      end;
    end
  else //RLE ?
    begin
      EID := DWORD(ImgData) + ImgInfo.Size;
      Pt := ImgInfo.Width;
      repeat
        Flag := (PByte(ImgData)^ and $80) = 0;
        Cnt := PByte(ImgData)^ and $7F;
        Inc(PChar(ImgData), 1);
        if Flag then //A repeating byte is set
          begin
            repeat
              if Cnt < Pt then
                begin
                  FillChar(P^, Cnt, PByte(ImgData)^);
                  Inc(PChar(P), Cnt);
                  Dec(Pt, Cnt);
                  Cnt := 0;
                end
              else
                begin
                  FillChar(P^, Pt, PByte(ImgData)^);
                  Dec(PChar(P), NW + ImgInfo.Width - Pt);
                  Dec(Cnt, Pt);
                  Pt := ImgInfo.Width;
                end
            until Cnt = 0;
            Inc(PChar(ImgData), 1);
          end
        else //  
          begin
            repeat
              if Cnt < Pt then
                begin
                  Move(ImgData^, P^, Cnt);
                  Inc(PChar(P), Cnt);
                  Inc(PChar(ImgData), Cnt);
                  Dec(Pt, Cnt);
                  Cnt := 0;
                end
              else
                begin
                  Move(ImgData^, P^, Pt);
                  Dec(PChar(P), NW + ImgInfo.Width - Pt);
                  Inc(PChar(ImgData), Pt);
                  Dec(Cnt, Pt);
                  Pt := ImgInfo.Width;
                end
            until Cnt = 0;
          end;
      until DWORD(ImgData) >= EID;
    end;
  //Adding 0 to NW (not really necessary)
  if NW <> ImgInfo.Width then
  begin
    P := Pointer(DWORD(Stream.Memory) + HSz + ImgInfo.Width);
    for i := 0 to ImgInfo.Height - 1 do
    begin
      FillChar(P^, NW - ImgInfo.Width, 0);
      Inc(PChar(P), NW);
    end;
  end;
  //flipping canvas vertically 
  if Mirror then
  begin
    P := Pointer(DWORD(Stream.Memory) + HSz);
    for i := 0 to ImgInfo.Height - 1 do
    begin
      for W := 0 to ImgInfo.Width div 2 - 1 do
      begin
        Tmp := PChar(P)[W];
        PChar(P)[W] := PChar(P)[ImgInfo.Width - W - 1];
        PChar(P)[ImgInfo.Width - W - 1] := Tmp;
      end;
      Inc(PChar(P), NW);
    end;
  end;
  Stream.Position := 0;
end;

destructor TArtInfo.Destroy;
begin
  Clear;
  inherited;
end;

function TArtInfo.GetImageInfoPtr(Picture, Frame: Integer): PImageInfo;
begin
  Result := @ImageInfo[Picture * FFrameCount + Frame];
end;

function TArtInfo.GetImageInfoPtr(Index: Integer): PImageInfo;
begin
  Result := @ImageInfo[Index];
end;

function TArtInfo.GetImagePtr(Picture, Frame: Integer): Pointer;
var
  i: integer;
begin
  Result := Images;
  for i := 0 to Picture * FFrameCount + Frame - 1 do
    Inc(PChar(Result), ImageInfo[i].Size);
end;

function TArtInfo.GetImagePtr(Index: Integer): Pointer;
var
  i: integer;
begin
  Result := Images;
  for i := 0 to Index - 1 do
    Inc(PChar(Result), ImageInfo[i].Size);
end;

function TArtInfo.IsEmpty: boolean;
begin
  Result := FStream = nil;
end;

procedure TArtInfo.LoadFile(const FileName: string);
begin
  Clear;
  if not SameText(ExtractFileExt(FileName), '.art') then Exit;
  if not FileExists(FileName) then Exit;
  FStream := TMemoryStream.Create;
  FStream.LoadFromFile(FileName);
  LoadInfo;
end;

procedure TArtInfo.LoadInfo;
var
  i: integer;
  P: Pointer;
begin
  Head := FStream.Memory;
  if Head[6] <> 0 then FPaletteCount := 4
  else if Head[5] <> 0 then FPaletteCount := 3
  else if Head[4] <> 0 then FPaletteCount := 2
  else FPaletteCount := 1;
  if (Head[0] = 0) or (Head[0] = 2) then
    begin
      FPictureCount := 8;
      FFrameCount := Head[8];
    end
  else
    begin
      FPictureCount := Head[8];
      FFrameCount := 1;
    end;
  P := Head;
  Inc(PChar(P), SizeOf(THead));
  for i := 0 to FPaletteCount - 1 do
  begin
    Palette[i] := P;
    Inc(PChar(P), SizeOf(TPaletteInfo));
  end;
  ImageInfo := P;
  Inc(PChar(P), SizeOf(TImageInfo) * FPictureCount * FFrameCount);
  Images := P;
end;

procedure TArtInfo.ReplaceBitmap(ImgIndx, PalIndx: Integer;
  ReplacePalette: boolean; Stream: TMemoryStream);
var
  Flag: Byte;
  i: integer;
  HSz, SSz, NW, Sz: DWORD;
  ImgInfo: PImageInfo;
  Src, SrcC, SrcP, SrcE, P, ImgData: Pointer;
  Dst, NewStream: TMemoryStream;
begin
  if IsEmpty then
    raise Exception.Create('Art is empty');
  if (PalIndx < 0) or (PalIndx >= PaletteCount) then PalIndx := 0;
  ImgInfo := GetImageInfoPtr(ImgIndx);
  NW := (ImgInfo.Width + 3) and $FFFFFFFC; //Making width multiple of 4
  HSz := SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader)
    + SizeOf(TPaletteInfo);
  SSz := HSz + ImgInfo.Height * NW;
  if Stream.Size - Stream.Position < Integer(SSz) then
    raise Exception.Create('Stream size too small');
  //TBitmapFileHeader
  P := Stream.Memory;
  Inc(PChar(P), Stream.Position);
  if (PBitmapFileHeader(P).bfType <> $4D42)
      or (PBitmapFileHeader(P).bfSize <> SSz)
      or (PBitmapFileHeader(P).bfOffBits <> HSz) then
    raise Exception.Create('Incorrect Image header');
  //TBitmapInfoHeader
  Inc(PChar(P), SizeOf(TBitmapFileHeader));
  if (PBitmapInfoHeader(P).biSize <> SizeOf(TBitmapInfoHeader)) then
    raise Exception.Create('Incorrect Image header');
  if (DWORD(PBitmapInfoHeader(P).biWidth) <> ImgInfo.Width)
      or (DWORD(PBitmapInfoHeader(P).biHeight) <> ImgInfo.Height) then
    raise Exception.Create('Incorrect Image Size');
  if (PBitmapInfoHeader(P).biPlanes <> 1)
      or (PBitmapInfoHeader(P).biBitCount <> 8) then
    raise Exception.Create('Incorrect Image format');
  //Palette
  Inc(PChar(P), SizeOf(TBitmapInfoHeader));
  if ReplacePalette then
    Move(P^, Palette[PalIndx]^, SizeOf(TPaletteInfo));
  //Image
  // We should remember that BMP files have the reverse line order
  // and the length of each line is multiple of 4 (the rest is ignored)
  Dst := TMemoryStream.Create;
  try
    Inc(PChar(P), SizeOf(TPaletteInfo) + (ImgInfo.Height - 1) * NW);
  //!!! There's a suspicion that it's incorrect to identify a data type this way
  // (Need to look for a flag)
    if ImgInfo.Width * ImgInfo.Height = ImgInfo.Size then //RAW ?
      begin
        for i := 0 to ImgInfo.Height - 1 do
        begin
          Dst.Write(P^, ImgInfo.Width);
          Dec(PChar(P), NW);
        end;
      end
    else //RLE ?
      begin
        GetMem(Src, ImgInfo.Width * ImgInfo.Height);
        try
          SrcP := Src;
          for i := 0 to ImgInfo.Height - 1 do
          begin
            Move(P^, SrcP^, ImgInfo.Width);
            Dec(PChar(P), NW);
            Inc(PChar(SrcP), ImgInfo.Width);
          end;
          SrcP := Src;
          SrcE := Pointer(DWORD(Src) + ImgInfo.Width * ImgInfo.Height - 1);
          while DWORD(SrcP) <= DWORD(SrcE) do
          begin
            SrcC := Pointer(DWORD(SrcP) + 1);
            while (DWORD(SrcC) <= DWORD(SrcE)) and (Byte(SrcC^) = Byte(SrcP^)) do
              Inc(PChar(SrcC));
            if DWORD(SrcC) > DWORD(SrcP) + 1 then //A sequence of repeating bytes
              begin
                //A sequence can't be longer than $7F
                if DWORD(SrcC) - DWORD(SrcP) > $7F then
                  SrcC := Pointer(DWORD(SrcP) + $7F);
                Flag := Byte(DWORD(SrcC) - DWORD(SrcP));
                Dst.Write(Flag, 1);
                Dst.Write(SrcP^, 1);
              end
            else //A sequence of _non_ repeating bytes
              begin
                while (DWORD(SrcC) < DWORD(SrcE)) and (Byte(SrcC^) <> Byte(Pointer(DWORD(SrcC) + 1)^)) do
                  Inc(PChar(SrcC));
                if (DWORD(SrcC) = DWORD(SrcE)) then
                  Inc(PChar(SrcC));
                //A sequence can't be longer than $7F
                if DWORD(SrcC) - DWORD(SrcP) > $7F then
                  SrcC := Pointer(DWORD(SrcP) + $7F);
                Flag := $80 or Byte(DWORD(SrcC) - DWORD(SrcP));
                Dst.Write(Flag, 1);
                Dst.Write(SrcP^, Flag and $7F);
              end;
            SrcP := SrcC;
          end;
        finally
          FreeMem(Src);
        end;
      end;
    //Writing new data to art stream
    if Integer(ImgInfo.Size) <> Dst.Size then
    begin
      NewStream := TMemoryStream.Create;
      try
        NewStream.Size := FStream.Size + Dst.Size - Integer(ImgInfo.Size);
        ImgData := GetImagePtr(ImgIndx);
        Sz := DWORD(ImgData) - DWORD(FStream.Memory);
        Move(FStream.Memory^, NewStream.Memory^, Sz);
        Inc(PChar(ImgData), ImgInfo.Size);
        P := Pointer(DWORD(NewStream.Memory) + Sz + DWORD(Dst.Size));
        Sz := DWORD(FStream.Size) - (DWORD(ImgData) - DWORD(FStream.Memory));
        Move(ImgData^, P^, Sz);
      except
        NewStream.Free;
        raise;
      end;
      FStream.Free;
      FStream := NewStream;
      LoadInfo;
      //Writing new size
      ImgInfo := GetImageInfoPtr(ImgIndx);
      ImgInfo.Size := Dst.Size;
    end;
    //Writing new data
    ImgData := GetImagePtr(ImgIndx);
    Move(Dst.Memory^, ImgData^, Dst.Size);
  finally
    Dst.Free;
  end;
end;

procedure TArtInfo.SaveFile(const FileName: string);
begin
  if IsEmpty then
    raise Exception.Create('Art is empty');
  FStream.SaveToFile(FileName);
end;

end.
